/*
 * Copyright (c) 2000-2004 Silicon Graphics, Inc All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * Contact information:  Silicon Graphics, Inc., 1500 Crittenden Lane,
 * Mountain View, CA  94043, or:
 *
 * http://www.sgi.com
 */

#include <stdarg.h>
#include <unistd.h>
#include <pwd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/syslog.h>

#define PAM_SM_ACCOUNT
#define PAM_SM_SESSION 

#include <security/_pam_macros.h>
#include <security/pam_modules.h>

#ifndef LINUX_PAM 
#include <security/pam_appl.h>
#endif

#include <jobctl.h>


static int disabled = 0;


/*
 * PAM framework looks for these entry-points to pass control to the 
 * account management module.
 */

/* 
 *  * Not all PAMified apps invoke session management modules.  So, we supply
 *   * this account management function for such cases.  Whenever possible, it
 *    * is still bet to use the session management version.
 *     */
PAM_EXTERN int pam_sm_acct_mgmt(pam_handle_t *pamh, int flags, int argc,
                const char **argv)
{
        int             retcode = PAM_SUCCESS;
        struct passwd   *passwd = NULL;
        char            *username = NULL;
        char            *service = NULL;
        struct job_create    jcreate;

        /* Get the username of the user associated with this job */
        retcode = pam_get_item(pamh, PAM_USER, (void *) &username);
        if (username == NULL || retcode != PAM_SUCCESS) {
                syslog(LOG_CRIT, "open_session - error recovering username");
                return PAM_AUTH_ERR;
        }

        /* Get the uid value for the user associated with this job */
        passwd = getpwnam(username);
        if (!passwd) {
                syslog(LOG_CRIT, "open_session - error getting passwd entry");
                return PAM_AUTH_ERR;
        }


        /* Get the service used to create this job */
        retcode = pam_get_item(pamh, PAM_SERVICE, (void *) &service);
        if (service == NULL || retcode != PAM_SUCCESS) {
                syslog(LOG_CRIT, "open_session - error recovering service");

                return PAM_AUTH_ERR;
        }

#if 0   /* this allowed root logins without creating jobs */

        if (passwd->pw_uid == 0) {
                syslog(LOG_INFO , "(%s) POE(pid=%d): user(uid=%d),"
                                " skipping job creation for superuser\n",
                                service,
                                getpid(),
                                passwd->pw_uid);
                disabled = 1;
                return PAM_SUCCESS;
        }       
#endif  


        /* Set up request structure to create job */
        jcreate.r_jid = 0;      /* return job ID for created job */
        jcreate.jid = 0;        /* requested job ID, 0 = no req. value */
        jcreate.user = passwd->pw_uid; /* uid associated with job */

        /* Request creation of new job */
        if (jobctl(JOB_CREATE, (void *)&jcreate) != 0) {
                /* Record failure to create job & session will fail */
                syslog(LOG_INFO, "jobctl(JOB_CREATE, ...): %s",
                                strerror(errno));
                return PAM_AUTH_ERR;
        }

	/* Record status of creating new job */
	if (jcreate.r_jid == 0) {
		/* Job creation was disabled */
		syslog(LOG_INFO, "(%s) POE(pid=%d): job creation disabled\n",
				service, getpid());
		disabled = 1;
	} else {
		/* New job successfully created */
        	syslog(LOG_INFO, "(%s) POE(pid=%d): job(jid=%0#18Lx) created for user %s(uid=%d) - using account management, no POE exit message will appear\n",
                        	service, getpid(), 
				(unsigned long long)jcreate.r_jid,
                        	username, passwd->pw_uid);
	}

        return PAM_SUCCESS;
}





/*
 * PAM framework looks for these entry-points to pass control to the
 * session module.
 */
 

PAM_EXTERN int pam_sm_open_session(pam_handle_t *pamh, int flags, int argc, 
		const char **argv)
{
	int 		retcode = PAM_SUCCESS;
	struct passwd	*passwd = NULL;
	char 		*username = NULL;
	char 		*service = NULL;
	struct job_create 	jcreate;

	/* Get the username of the user associated with this job */
	retcode = pam_get_item(pamh, PAM_USER, (void *) &username); 
	if (username == NULL || retcode != PAM_SUCCESS) {
		syslog(LOG_CRIT, "open_session - error recovering username");
		return PAM_SESSION_ERR;
	}

	/* Get the uid value for the user associated with this job */
	passwd = getpwnam(username);
	if (!passwd) {
		syslog(LOG_CRIT, "open_session - error getting passwd entry");
		return PAM_SESSION_ERR;
	}


	/* Get the service used to create this job */	
	retcode = pam_get_item(pamh, PAM_SERVICE, (void *) &service);
	if (service == NULL || retcode != PAM_SUCCESS) {
		syslog(LOG_CRIT, "open_session - error recovering service");
		return PAM_SESSION_ERR;
	}

#if 0

	if (passwd->pw_uid == 0) {
		syslog(LOG_INFO , "(%s) POE(pid=%d): user(uid=%d),"
				" skipping job creation for superuser\n",
				service,
				getpid(),
				passwd->pw_uid);
		disabled = 1;
		return PAM_SUCCESS;
	}	
#endif	


	/* Set up request structure to create job */
	jcreate.r_jid = 0;	/* return job ID for created job */
	jcreate.jid = 0;	/* requested job ID, 0 = no req. value */
	jcreate.user = passwd->pw_uid; /* uid associated with job */

	/* Request creation of new job */
	if (jobctl(JOB_CREATE, (void *)&jcreate) != 0) {
		/* Record failure to create job & session will fail */
		syslog(LOG_INFO, "jobctl(JOB_CREATE, ...): %s",
				strerror(errno));
		return PAM_SESSION_ERR;
	} 

	/* Record status of creating new job */
	if (jcreate.r_jid == 0) {
		/* Job creation was disabled */
		syslog(LOG_INFO, "(%s) POE(pid=%d): job creation disabled\n",
				service, getpid());
		disabled = 1;
	} else {
		/* New job successfully created */
        	syslog(LOG_INFO, "(%s) POE(pid=%d): job(jid=%0#18Lx) created for user %s(uid=%d) - using session management, a POE exit message will appear\n",
                        	service, getpid(), 
				(unsigned long long)jcreate.r_jid,
                        	username, passwd->pw_uid);
	}

	return PAM_SUCCESS; 
}


PAM_EXTERN int pam_sm_close_session(pam_handle_t *pamh, int flags, int argc, 
		const char **argv)
{
	int 		retcode = PAM_SUCCESS;
	char 		*username = NULL;
	char 		*service = NULL;
	struct job_getjid	getjid;

	/* Get the username of the user associated with this job */
	retcode = pam_get_item(pamh, PAM_USER, (void *) &username);
	if (username == NULL || retcode != PAM_SUCCESS) {
		syslog(LOG_CRIT, "open_session - error recovering username");
		return PAM_SESSION_ERR;
	}

	/* Get the service used to create this job */	
	retcode = pam_get_item(pamh, PAM_SERVICE, (void *) &service);
	if (service == NULL || retcode != PAM_SUCCESS) {
		syslog(LOG_CRIT, "open_session - error recovering service");
		return PAM_SESSION_ERR;
	}

	getjid.r_jid = 0;
	getjid.pid = getpid();

	if (!disabled) {
		if (jobctl(JOB_GETJID, (void *)&getjid) != 0) {
			syslog(LOG_CRIT, "jobctl - error getting job ID");
		}
	}

	if (getjid.r_jid > 0) {
		syslog(LOG_INFO, "(%s) POE(pid=%d): job(jid=%0#18Lx) POE process is exiting\n", 
				service, getpid(), 
				(unsigned long long)getjid.r_jid);
	} else {
		syslog(LOG_INFO, "(%s) POE(pid=%d): POE process is exiting. Not attached to any job\n",
				service, getpid());
	}

	return PAM_SUCCESS;
}




